/*
 * MarcAdd - a program to load MARC records into PYTHEAS database structure.
 *
 * 		Art Rhyno (arhyno@uwindsor.ca), June 1999
 *		(c) Copyright GNU General Public License (GPL)
 *
 * 		Revised:
 *			Aug. 2001 - changed for authorities
 *			Jan. 2002 - added command line options
 *			Apr. 2002 - changed heading/authorities structure
 *			Sept. 2002 - added initial Topic Map setup
 *
 */

package pytheas;

//PYTHEAS
import pytheas.*;

//Standard Java
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.util.*;
import java.io.*;
import java.io.PrintWriter;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.DriverManager;

//SAX
import org.xml.sax.ContentHandler;
import org.apache.xml.serialize.*;

//Castor
import org.exolab.castor.jdo.JDO;
import org.exolab.castor.jdo.Database;
import org.exolab.castor.jdo.OQLQuery;
import org.exolab.castor.jdo.QueryResults;
import org.exolab.castor.util.Logger;
import org.exolab.castor.xml.Marshaller;
import org.exolab.castor.mapping.Mapping;

public class MarcAdd  {

	// All of these should be put in their own XML file

	public static final String DatabaseFile = "database.xml";
	public static final String MappingFile = "mapping.xml";

	private Mapping  _mapping;
	private JDO      _jdo;

	private static String endField = 	"\u001e";	// MARC 1E
	private static String subRecord = 	"\u001d";	// MARC 1D
	private static String subField = 	"\u001f";	// MARC 1F
	private static String Identifier = 	"Resource Identifier";

    	private static String nameAssociationVerb = "written-by-";
    	private static String subjectAssociationVerb = "is-about-";

    	private static int identMaxLen = 20;
  
    	public static void main ( String args[] )
		throws ClassNotFoundException, IOException, InstantiationException, 
		IllegalAccessException, Exception
    	{	    
		PrintWriter writer;
		FileOutputStream out = null;

		String input;
		String user = "";
		String password = "";
		String db = "";
		String host = "";
		String port = "";
		String holdingsFields = "";

		// preset these for now, lazy I know....
		String theUrl = "http://localhost?pyserv=";
		String inputFile = "MarcAdd.mrc";
		int maxFieldLen = 255;
		boolean authoritiesOnly = false;
		int j = args.length;

		for (int i=0;i<j;i++) {
			if (args[i].equals("-i") && i+1 < j)
				inputFile = args[++i];
			if (args[i].equals("-a"))
				authoritiesOnly = true;
			if (args[i].equals("-l") && i+1 < j)
				maxFieldLen = verifyNum(args[++i]);
			if (args[i].equals("-?")) {
				System.out.println("MarcAdd [-i input file] " +
					"[-a] [-l maximum length] [-?]");
				System.exit(1);
			}//if args
		} // for i

  		System.out.println("Starting, please wait...");

		DataInputStream D1 = new DataInputStream(new FileInputStream("MarcAdd.ini"));
		BufferedReader In1 = new BufferedReader(new InputStreamReader(D1)); 
    
		while ((input = In1.readLine()) != null) {
	    		String parm = "";
	    		String parmValue = "";
	    		StringTokenizer tokens = new StringTokenizer(input,"=");

	    		while (tokens.hasMoreTokens()) {
				parm = tokens.nextToken().trim();
				if (parmValue.equals("holdings"))
		    			holdingsFields = parm;
				else if (parmValue.equals("url"))
		    			theUrl = parm;
				else if (parmValue.equals("crosswalk"))
		    			CrosswalkList.Add(new CrosswalkEntry(parm));
				parmValue = parm;		    
	    		}//while tokens
		}//while input
	
		//check for holdings
		if (holdingsFields.length() == 0) 
		{
	    		System.err.println("Missing parameter(s) in MarcAdd.ini");
	    		System.exit(-1);
		}//if holdingsFields
	
		// Clear log
		PrintWriter clearit = new PrintWriter(new FileOutputStream("MarcAdd.log"));
		clearit.close();
		out = new FileOutputStream("pytheas.log");

		writer = new Logger(out).setPrefix("pytheas");

		MarcAdd m = new MarcAdd(holdingsFields,writer,inputFile,authoritiesOnly,maxFieldLen,theUrl);
	}//main
    
    	// MarcAdd - process specified file into database
    	public MarcAdd(String holdingsFields, PrintWriter writer, String inputFile, 
		boolean authoritiesOnly, int maxFieldLen, String theUrl) throws Exception
    	{
		Database	db = null;

		// use lots of vectors, isn't as necessary now that "many to many" is
		// much more stable in Castor, but haven't got around to changing yet
		Vector allNameHeadings = new Vector();
		Vector allJdoNameHeadings = new Vector();
		Vector trackJdoNameHeadings = new Vector();
		Vector allSubjectHeadings = new Vector();
		Vector allJdoSubjectHeadings = new Vector();
		Vector trackJdoSubjectHeadings = new Vector();

        	_mapping = new Mapping( getClass().getClassLoader() );
        	_mapping.setLogWriter( writer );
        	_mapping.loadMapping( getClass().getResource( MappingFile ) );

        	_jdo = new JDO();
        	_jdo.setLogWriter( writer );
        	_jdo.setConfiguration( getClass().getResource( DatabaseFile ).toString() );
        	_jdo.setDatabaseName( "test" );
	
		db = _jdo.getDatabase();

		// ok, now on to our main objects
		Bib bib;
		Resource resource;
		Topicmap topicMap;

		String recordIdent;
		char bibArray[];

		// some utility values
		int bibFldId, bibFieldLen, bibCurrentPos, bibBegData;
		int bibRecLen = 0;
		int offset = 0;
		int headerLen = 0;
		int recordCnt = 0;
   
		FileInputStream in = new FileInputStream(inputFile);
		BufferedInputStream bin = new BufferedInputStream(in);
	
		// work through MARC input file
		if (!authoritiesOnly) {

			// for testing, you would limit the processed records as shown:
			// while (bin.available() > 0 && recordCnt < 5) {

			while (bin.available() > 0) {
				db.begin();

				bib = new Bib();
				resource = new Resource();
				topicMap = new Topicmap();
			
				recordIdent = "";
				bibArray = readMarc(bin,5);
				bibRecLen = verifyNum(String.valueOf(bibArray,0,5)) - 5;
				offset = 0;	    
				bibArray = readMarc(bin,bibRecLen);
				headerLen = verifyNum(String.valueOf(bibArray,7,5)) - 5;
				bibBegData = headerLen;
				bibCurrentPos = 19;

				while (bibCurrentPos < bibBegData - 1) {
					// typical MARC processing, should replace with MARC util
					bibFldId = Integer.parseInt(String.valueOf(bibArray,
						bibCurrentPos,3));
					bibCurrentPos += 3;
					bibFieldLen = Integer.parseInt(String.valueOf(bibArray,
						bibCurrentPos,4));
					bibCurrentPos += 4;
					offset = Integer.parseInt(String.valueOf(bibArray,bibCurrentPos,
						5));
					bibCurrentPos += 5;

					if (bibFldId  == 1) {
						recordIdent = String.valueOf(bibArray,offset + 
							bibBegData,bibFieldLen - 1);		    
						resource.setUri(theUrl + recordIdent);
						db.create(resource);

						/* Just create a basic TopicMap for now, this will beed to
						   be fleshed out. */
						topicMap.setTopicMapIdent(recordIdent);
						db.create(topicMap);

						// do inital bib values
            					bib.setResource( resource );
						bib.setRecordStatus(String.valueOf(bibArray,0,1));
						bib.setRecordType(String.valueOf(bibArray,1,1));
						bib.setEncodingLevel(String.valueOf(bibArray,12,1));
						bib.setDesCatForm(String.valueOf(bibArray,13,1));
					}//if

					if (!processValue(db,resource,topicMap,bib,bibFldId,fixField(
						String.valueOf(bibArray,offset + bibBegData,bibFieldLen)),
						recordCnt,writer,allNameHeadings,allJdoNameHeadings,
						allSubjectHeadings,allJdoSubjectHeadings,maxFieldLen)) 
					{
		    				System.err.println("Problem at record: " + recordIdent +
							"sequence#: " + recordCnt);
		    				System.exit(-1);
					}//if !processValue
				} // while bibCurrentPos

				db.create(bib);

				if (headingsResourceTrack(resource.getId(),topicMap.getId(),allNameHeadings,
					allJdoNameHeadings,trackJdoNameHeadings) && 
					headingsResourceTrack(resource.getId(),topicMap.getId(),
					allSubjectHeadings,allJdoSubjectHeadings,trackJdoSubjectHeadings)) 
				{
					db.commit();
				}//if headingsResourceTrack
	
				allJdoNameHeadings = new Vector();
				allJdoSubjectHeadings = new Vector();

				recordCnt++;
				System.out.print("." + recordCnt);
			} //while bin.available
		} //if !authoritiesOnly
	
		db.close();
		int newHeadingCnt = updateHeadings(db,trackJdoNameHeadings,
			"SELECT n FROM pytheas.Names n WHERE name = \"",
			"SELECT a FROM pytheas.Associations a WHERE topicIdent = \"" + nameAssociationVerb,
			true,theUrl,maxFieldLen) +
			updateHeadings(db,trackJdoSubjectHeadings,
			"SELECT s FROM pytheas.Subjects s WHERE subject = \"",
			"SELECT a FROM pytheas.Associations a WHERE topicIdent = \"" + subjectAssociationVerb,
			false,theUrl,maxFieldLen);


		System.out.println ("\nProcessing Complete: " + recordCnt + 
			" record(s) processed.");
	}//MarcAdd
        
    	public boolean headingsResourceTrack(int resourceId,int topicMapId, Vector allHeadings,Vector allJdoHeadings, 
		Vector trackJdoHeadings) 
    	{
		if (allJdoHeadings.size() > 0) {
			for (int i=0; i<allJdoHeadings.size();i++) {
				Integer theNo = (Integer)allJdoHeadings.elementAt(i);
				int headingNo = theNo.intValue();
				if (headingNo >= trackJdoHeadings.size()) {
					trackJdoHeadings.addElement(
						new HeadingEntry(
						(String)allHeadings.elementAt(headingNo),
						resourceId,topicMapId));
				} else {
					HeadingEntry theEntry = (HeadingEntry) 
						trackJdoHeadings.elementAt
						(headingNo);
					theEntry.addResource(resourceId);
					theEntry.addTopicMap(topicMapId);
				}//if
			}//for
		}//if

		return true;
	}//headingsBibTrack


	/* 	updateHeadings currently sets up Topic Map associations in the same way that
		authorities bring records together. This is flimsy at best, and far short of
		what Topic Maps can represent, but it's a start.
	*/
	public int updateHeadings(Database db,Vector trackJdoHeadings, String headingOqlStmt,
		String topicOqlStmt, boolean isName, String theUrl, int maxFieldLen) 
		throws Exception
    	{

		Names	 	name = null;
		Subjects	subject = null;
		Associations	association = null;
		Topicmap	thisTopicMap = null;
		Topic		thisTopic=null;
		Resource	thisResource = null;
		OQLQuery	resourceOql,topicOql,headingOql;
		QueryResults  	results;

		Collection 	associations, names, subjects, resources, topics;
		String 		theValue;

		System.out.print("\nNow working on Headings...");

		db = _jdo.getDatabase();
		db.begin();

		int lastResource = -1;
		int lastTopicMap = -1;
		int newHeadings = 0;

		// cycle through authority headings
		for (int i=0; i<trackJdoHeadings.size();i++) {	
			HeadingEntry theEntry = (HeadingEntry) 
				trackJdoHeadings.elementAt
				(i);
			String theHeading = theEntry.returnHeading();
			Vector resourceIds = theEntry.returnResources();
			Vector topicMapIds = theEntry.returnTopicMaps();
        		headingOql = db.getOQLQuery( headingOqlStmt + 
				theHeading + "\"" );
        		results = headingOql.execute();
		
        		if ( ! results.hasMore() ) {
				logEvent( "Heading not found: " + theHeading );

				if (isName) {
					name = new Names();
					name.setName(theHeading);
					name.setType("n");
					name.setUri(theUrl + "?name=" + theHeading);
					name.setIsCrossref(false);
					name.setIsEstablished(false);
					db.create(name);	
				} else {
					subject = new Subjects();
					subject.setSubject(theHeading);
					subject.setType("s");
					subject.setUri(theUrl + "?subject=" + theHeading);
					subject.setIsCrossref(false);
					subject.setIsEstablished(false);
					db.create(subject);
				}//if isName

				resources = new ArrayList();
				newHeadings++;
			} else {
				if (isName) {
					name = (Names) results.next();
					resources = name.getResources();
				} else {
					subject = (Subjects) results.next();
					resources = subject.getResources();
				}//if isName
			}//if !results.hasMore

 
			// check to see if association exists for heading
	       		topicOql = db.getOQLQuery( topicOqlStmt + 
				theHeading + "\"" );

			results = topicOql.execute();
		
        		if ( ! results.hasMore() ) {
				logEvent( "Topic not found: " + theHeading );
				association = new Associations();
				if (isName) 
					theValue = nameAssociationVerb + theHeading;
				else
					theValue = subjectAssociationVerb + theHeading;
				if (theValue.length() > maxFieldLen) 
					theValue = theValue.substring(0,
						maxFieldLen - 1);
				association.setTopicIdent(theValue);
				db.create(association);
				topics = new ArrayList();
        		} else {
				association = (Associations) results.next();
				topics = association.getTopics();
			}//if !results.hasMore

			// work with resources that have been created or updated
			for (int j=0;j<resourceIds.size();j++) {
				thisResource = (Resource) db.load(Resource.class,
					resourceIds.elementAt(j)); 
				Integer theNo = (Integer) resourceIds.elementAt(j);
				int nowRes = theNo.intValue();
				logEvent("working with " + nowRes);

				if (lastResource != nowRes) {
					resourceOql = db.getOQLQuery( 
						"SELECT r FROM pytheas.Resource r WHERE id=$1");
					resourceOql.bind(resourceIds.elementAt(j));
					results = resourceOql.execute();
					if ( results.hasMore() ) 
						thisResource = (Resource) results.next();
				}//if lastResource != nowRes

				resources.add(thisResource);
				if (isName) {
					names = thisResource.getNames();
					if (names == null)
						names = new ArrayList();
					name.setResources(resources);
					names.add(name);
					thisResource.setNames(names);
				} else {
					subjects = thisResource.getSubjects();
					if (subjects == null)
						subjects = new ArrayList();
					subject.setResources(resources);
					subjects.add(name);
					thisResource.setSubjects(subjects);
				}//if isName

				lastResource = nowRes;			
				System.out.print(".");
			}// for j

			// cycle through Topic Maps
			for (int j=0;j<topicMapIds.size();j++) {
				thisTopicMap = (Topicmap) db.load(Topicmap.class,
					topicMapIds.elementAt(j)); 
				Integer theNo = (Integer) topicMapIds.elementAt(j);
				int nowTopicMap = theNo.intValue();
				logEvent("working with " + nowTopicMap);

				if (lastTopicMap != nowTopicMap) {
					topicOql = db.getOQLQuery( 
						"SELECT t FROM pytheas.Topicmap t WHERE id=$1");
					topicOql.bind(topicMapIds.elementAt(j));
					results = topicOql.execute();

					if ( results.hasMore() ) {
						thisTopicMap = (Topicmap) results.next();

						/*
						cheat here because we are only creating one topic,
						eventually will want to specify "title" or whatever
						*/

						topics = thisTopicMap.getTopics();
						Iterator itor = topics.iterator();
						if (itor.hasNext()) {
							thisTopic = (Topic) itor.next();
							logEvent("topic is " + thisTopic);
						} else
							logEvent("no topic");
					}//if results
				}//if lastTopicMap != nowTopicMap

				if (thisTopic != null) {
					topics.add(thisTopic);
					associations = thisTopic.getAssociations();
					if (associations == null)
						associations = new ArrayList();
					association.setTopics(topics);
					associations.add(association);
					thisTopic.setAssociations(associations);
				}//if thisTopic

				lastTopicMap = nowTopicMap;			
				System.out.print(".");
			}// for j
		}// for i

		//don't need to commit if we haven't done anything
		if (trackJdoHeadings.size() > 0) 
			db.commit();;

		logEvent("Headings Processed: " + trackJdoHeadings.size());
		db.close();	

		return newHeadings;
	} //updateHeadings

    	// fillValue- return string with inserted value placed in front
	public static final String fillValue (int inNum, String filler, int fillNum) 
    	{
		StringBuffer numHolder;
		int i, valueLen;
	
		numHolder = new StringBuffer(fillNum * 2);
		numHolder.append(inNum);
		if (numHolder.length() < fillNum) {
	    		valueLen = fillNum - numHolder.length();
	    		for (i = 0; i < valueLen; i++)
				numHolder.insert(0,filler);
		}// if numHolder

		return numHolder.toString();
    	}//fillValue
    
    	//logEvent - append specified event to log
    	public static boolean logEvent(String eventMessage)
    		throws IOException
    	{
		// Always append
		PrintWriter output =            
			new PrintWriter(new FileOutputStream("MarcAdd.log",true));      
	
		output.println(eventMessage);
		output.close();
		
		return true;
    	}//logError    
    
    	//verifyNum - make sure numbers convert properly
    	public static int verifyNum(String valueToConvert)
    	{
		int i;
	
		try { i = Integer.parseInt(valueToConvert); }
		catch (NumberFormatException e) {
	    		return -1;
		}//catch NumberFormatException

		return i;
	}//verifyNum
    
    	//readMarc - assemble the total MARC record
    	public static char[] readMarc(BufferedInputStream bin, int recLen) 
		throws IOException
    	{
		int offset = 0;
		int totLen = 0;
		int nBytesRead = 0;

		String marcRec;

		byte bufferIn[] = new byte[recLen];
		nBytesRead = bin.read(bufferIn);
	
		if (nBytesRead < recLen) 
	    		totLen = recLen - nBytesRead;

		while (totLen > 0) {
	    		byte chunk[] = new byte[totLen];
	    		offset = bin.read(chunk);
	    		totLen -= offset;
	    		System.arraycopy(chunk,0,bufferIn,nBytesRead,offset);  
	    		nBytesRead += offset;
		}// while totLen

		marcRec = new String (bufferIn);
		return marcRec.toCharArray();
    	}//readMarc
    
    	//fixField - 	change entries that contain quotation marks so database won't reject them
	//		this isn't a problem for MySQL, have to add logic to check target db engine
    	public String fixField(String subfValue)
    	{
		String pieceSubf;	
		StringBuffer fixedSubf = new StringBuffer();
		int i = 0;
	
		StringTokenizer tokens = new StringTokenizer(subfValue,"\'");
		while (tokens.hasMoreTokens()) {
	    		pieceSubf = tokens.nextToken();
	    		fixedSubf.append(pieceSubf);
	    		i += pieceSubf.length();

	    		if (i != subfValue.length()) {
				fixedSubf.append("\\\"");
				i++;
	    		}// if i
		}//while tokens

		return fixedSubf.toString();
	}//fixField
    
    	//processValue - this is where the field associated with the tag is extracted
	public boolean processValue(Database db, Resource resource, Topicmap topicMap, Bib bib, 
		int tagNum, String marcField, int recordID, PrintWriter writer, Vector allNameHeadings,
		Vector allJdoNameHeadings, Vector allSubjectHeadings, Vector allJdoSubjectHeadings, 
		int maxFieldLen) throws Exception    
    	{
		Rdf		rdf = null;
        	MarcVarTag 	tag=null;

		String 		fieldValue, ind1,ind2, marker, subfValue, rawValue;
		int 		i = 0;
		boolean 	getIndicators = true;
	
		ind1 = ind2 = " ";
	
		// 001 - Record Number
		if (tagNum == 1) { 
	    		i = marcField.indexOf(endField);

	    		if (i > 0)
				fieldValue = marcField.substring(0,i);
	    		else
				fieldValue = marcField;

	    		bib.setOrigNo(fieldValue);
		}//if 001
	
		// 008 - fixed fields
		if (tagNum == 8) {
			bib.setTypeDate( marcField.substring(6,7) );
			bib.setBegPubDate( marcField.substring(7,11) );
			bib.setEndPubDate( marcField.substring(11,15) );
			bib.setPubPlace( marcField.substring(15,18) );
			bib.setIllCode( marcField.substring(18,22) );
			bib.setTargetAud( marcField.substring(22,23) );
			bib.setItemForm( marcField.substring(23,24) );
			bib.setContNature( marcField.substring(24,28) );
			bib.setGovCode( marcField.substring(28,29) );
			bib.setConfCode( marcField.substring(29,30) );
			bib.setFestschrift( marcField.substring(30,31) );
			bib.setOwnIndex( marcField.substring(32,32) );
			bib.setFiction( marcField.substring(32,34) );
			bib.setBiography( marcField.substring(34,35) );
			bib.setLangCode( marcField.substring(35,38) );
		}//if 008

		// and everything else
		if (tagNum == 7 || tagNum >= 10) {
			StringTokenizer tokens = new StringTokenizer(marcField,subField);

	    		while (tokens.hasMoreTokens()) {
				rawValue = tokens.nextToken();
				i = rawValue.indexOf(endField);
				if (i > 0)
		    			fieldValue = rawValue.substring(0,i);
				else
		    			fieldValue = rawValue;
				if (getIndicators) {
		    			ind1 = fieldValue.substring(0,1);
		    			ind2 = fieldValue.substring(1,2);

		    			tag = new MarcVarTag();
		    			tag.setTagName( fillValue(tagNum,"0",3) );
	    	    			tag.setInd1( ind1 );
	            			tag.setInd2( ind2 );
					tag.setMarc(bib);
		    			getIndicators = false;		    
				} else {
		    			marker = fieldValue.substring(0,1);
		    			subfValue = fieldValue.substring(1);

		    			if (!dbUpdate(db,bib,tag,tagNum,ind1,ind2,marker,subfValue,
						writer,allNameHeadings,allJdoNameHeadings,allSubjectHeadings,
						allJdoSubjectHeadings,maxFieldLen)) 
					{
						System.err.println("problem inserting tag: " + 
							fillValue(tagNum,"0",3) +
			    				" on subfield: " + marker);
						return false;
		    			} else {
						CrosswalkList sl = null;
						while ((sl = CrosswalkList.Next(sl)) != null)
						{
			    				CrosswalkEntry sn = sl.Data();		
			    				if (processRule(tagNum,sn)) {
								if (!dbCrosswalkSelect(db,resource,topicMap, 	
									sn,tagNum,ind1,ind2,marker,
									subfValue,writer,maxFieldLen)) 
								{
									System.err.println("problem " +
										"inserting schema " +
										"information for: " +
					    					fillValue(tagNum,"0",3) + 
										"on subfield: " + marker);
									return false;
								}//if !dbCrosswalkSelect
			    				}//if processRule
						}//while s1
		    			}//if !dbUpdate

		    			if (subfValue.length() > maxFieldLen)
						logEvent("Length: " + subfValue.length() + " ->" + subfValue);
				}//if getIndicators
	    		}//while tokens
		}//if tagNum

		if (tag != null) {
			logEvent("adding " + tag);
			Collection vartags = bib.getVartags();

			if (vartags == null)
				vartags = new ArrayList();
			vartags.add(tag);
			bib.setVartags( vartags );
		}//if tag
	
		return true;
	}//processMarc
    
	//processRule - store value if tag matches rule
    	public static boolean processRule(int tagNo, CrosswalkEntry schemaValue)
    	{
		boolean matchFlag = false;
		StringBuffer markers = new StringBuffer();
		String subFields;
	
		String matchValue = schemaValue.returnRule();
		int i = matchValue.indexOf(fillValue(tagNo,"0",3));

		if (i >= 0) {
	    		matchFlag = true;
	    		if (!schemaValue.ruleSet()) {
				int j = matchValue.indexOf("$");
				if (j > 4) 
		    			schemaValue.setInd(matchValue.substring(4,6));
				i = matchValue.indexOf("$");
				if (i > 0) {
		    			StringTokenizer tokens = new 
						StringTokenizer(matchValue.substring(i),"$");
		    			while (tokens.hasMoreTokens()) 
						markers.append(tokens.nextToken().trim());
		    			schemaValue.setSubf(markers.toString());
				}//if i
				schemaValue.setRule();
	    		}//if !schemaValue
		}//if i

		return matchFlag;
	}//processRule
    
	public boolean dealWithHeadings(String headingValue,
		Vector allHeadings,Vector allJdoHeadings) throws Exception
    	{    

		int headingNo = -1;

		headingNo = allHeadings.indexOf(headingValue);

		if (headingNo != -1) {
			allJdoHeadings.addElement(new Integer(headingNo));
		} else {
			allHeadings.addElement(headingValue);
			allJdoHeadings.addElement(new Integer(allHeadings.size() - 1));
		}//if

		return true;
	}//dealWithHeadings

	//tableSelect - decide where the values go and add them	    
    	public boolean dbUpdate(Database db, Bib bib, MarcVarTag tag, int tagNum, String ind1, 
		String ind2, String marker, String subfValue, PrintWriter writer, Vector allNameHeadings, 
		Vector allJdoNameHeadings, Vector allSubjectHeadings, Vector allJdoSubjectHeadings,
		int maxFieldLen) throws Exception
	{    

		MarcVarTagSubf subf;
		Collection subfs;

		int threshold = maxFieldLen - 1;

		if (subfValue != null )  {
			String theValue = subfValue;
			String headingValue = theValue;

			int i = subfValue.length();
			int j = 0;

			while ( i > 0) {		
				subf = new MarcVarTagSubf();
            			subf.setSeqNo( j++ );
				subf.setSubfCode( marker );

				if (i <= maxFieldLen) {
					subf.setSubfValue(theValue);
				} else {
					// break on a space, this needs more work...
					while (threshold >= 0 && subfValue.charAt(threshold) != ' ')
						--threshold;
					if (threshold <= 0)
						threshold = maxFieldLen - 1;

					subf.setSubfValue( theValue.substring(0,threshold));

					if (headingValue.length() > 0)
						headingValue = theValue.substring(0,threshold);
					theValue = theValue.substring(threshold + 1);
				}//if i

				if (tagNum == 100 && marker.equals("a") && headingValue.length() > 0) {
					tag.setControlFlag(true);
					// put junk in here for now, at some point this may point to an
					// authority somewhere else on the web
					tag.setControlUri("http://tag");
					headingValue = headingValue.toUpperCase();
					headingValue = headingValue.trim();

					if (!dealWithHeadings(normalizeEntry(headingValue),
						allNameHeadings,allJdoNameHeadings))
					{
						headingValue = "";
					}//if !dealWithHeadings
				}//if tagNum

				if (tagNum == 650 && marker.equals("a") && headingValue.length() > 0) {
					tag.setControlFlag(true);
					tag.setControlUri("http://tag");
					headingValue = headingValue.toUpperCase();
					headingValue = headingValue.trim();

					if (!dealWithHeadings(normalizeEntry(headingValue),
						allSubjectHeadings,allJdoSubjectHeadings))
					{
						headingValue = "";
					}//if !dealWithHeadings
				}//if tagNum

				i -= (threshold + 1);
				subfs = tag.getSubfs();

				if (subfs == null)
					subfs = new ArrayList();
				subf.setMarcvartag(tag);
				subfs.add(subf);
				tag.setSubfs(subfs);
			}//while
		}//if subfValue

		return true;
	}//tableSelect

	// this is primitive, would like to follow normalization rules, these are based on NACO
    	public String normalizeEntry(String abnormalEntry) 
    	{
		String theBadList = "[]';:<>/\"%_\\().,-+!@#$%^&*";
		String theCleanEntry = "";
	
		StringTokenizer tokens = new StringTokenizer(abnormalEntry,theBadList);

		while (tokens.hasMoreTokens()) 
			theCleanEntry = theCleanEntry + tokens.nextToken();

		return theCleanEntry;
    	}//normalizeEntry

    	// dbCrosswalkSelect - insert crosswalk values
    	public boolean dbCrosswalkSelect(Database db,Resource resource,Topicmap topicMap,
		CrosswalkEntry cn, int tagNum, 
		String ind1, String ind2, String marker, String subfValue, PrintWriter writer,
		int maxFieldLen) throws Exception
	{
		Rdf rdf;
		Topic topic;
		Basename basename;

		boolean isIdentifier = false;
	
		if (subfValue != null) {

	   		if (cn.isSubf()) {
				if (cn.returnSubf().indexOf(marker) == -1)
		    			return true;
	    		}//if cn.isSubf

			if (cn.isIndicator1()) {
				if (cn.returnInd1().indexOf(ind1) == -1)
		    			return true;
	    		}//if cn.isIndicator1

			if (cn.isIndicator2()) {
				if (cn.returnInd2().indexOf(ind2) == -1)
		    			return true;
	    		}//if cn.isIndicator2

			rdf = new Rdf();

			if (subfValue.indexOf("http://") != -1) 
				isIdentifier = true;

			rdf.setResource(resource);
			rdf.setIsType(false);
			rdf.setPredicate(cn.returnPredicate());
			// rdf.setNameSpace(cn.returnCrosswalk());
			// Hardcode DC for now <-- THIS NEEDS TO BE UPDATED
			rdf.setNameSpace("http://purl.org/dc/elements/1.1/");
			rdf.setNameCode("DC");

			String theValue = subfValue;
			if (theValue.length() > maxFieldLen)
				theValue = subfValue.substring(0,maxFieldLen - 1);
			rdf.setValue(theValue);

			if (!isIdentifier) {
				theValue = theValue.toUpperCase();
				theValue = theValue.trim();
				rdf.setNormalEntry(normalizeEntry(theValue));
			} else
				rdf.setNormalEntry(theValue);

			db.create(rdf);
			Collection rdfs = resource.getRdfs();
			if (rdfs == null)
				rdfs = new ArrayList();
			rdfs.add(rdf);
			resource.setRdfs( rdfs );

			if (cn.returnPredicate().indexOf("Title") != -1) {
				topic = new Topic();
				topic.setTopicMap(topicMap);
				if (theValue.length() > identMaxLen)
					theValue = theValue.substring(0,identMaxLen - 1);
				topic.setTopicIdent("#" + theValue);
				topic.setResourceRef(resource.getId());
				topic.setIsPublished(false);
				db.create(topic);

				basename = new Basename();
				basename.setTopic(topic);
				basename.setBaseNameString(theValue);
				db.create(basename);
			
				/* 
					this is an extra step in this case, but go this route
		   			for multiple values 
				*/

				Collection basenames = topic.getBasenames();
				if (basenames == null)
					basenames = new ArrayList();
				basenames.add(basename);
				topic.setBasenames(basenames);	

				Collection topics = topicMap.getTopics();
				if (topics == null)
					topics = new ArrayList();
				topics.add(topic);
				topicMap.setTopics(topics);
			}//if cn
			
		}// if subfValue

		return true;
	}//tableCrosswalkSelect            
} //MarcAdd
